import os
import sqlite3
import random
import logging
import uuid

DB_DIR = '/var/spool/kepi'
DB_FILENAME = 'kepi.sqlite3'

STATES = {
        'U': 'unverified',
        'I': 'inbox',
        'R': 'received',
        'O': 'outbox',
        'S': 'sent',
        'Z': 'gone',
        'X': 'errors',
        }

TABLE_CREATION = [
        (
            'create table message ('
            'uuid varchar(36), '
            'state varchar(1), '
            'content varchar'
            ')'
            ),
        ]

logger = logging.getLogger('kepi.daemon')

class Spool:
    def __init__(self,
                 path=DB_DIR,
                 ):

        def mkdir(path):
            try:
                os.makedirs(path)
            except FileExistsError:
                pass

        self.path = path

        self.db = sqlite3.connect(
                os.path.join(path, DB_FILENAME)
                )

        self._ensure_tables()

    def _db_exec(self, sql, args=(), cu=None, ignore_errors=False):
        if cu is None:
            cu = self.db.cursor()

        try:
            result = cu.execute(sql, args)
            logger.debug("> %s %s", sql, args)
        except sqlite3.OperationalError as e:
            if ignore_errors:
                pass
            else:
                logger.debug("> %s %s --> error: %s",
                             sql, args, e)
                raise

        return result

    def _ensure_tables(self):
        cu = self.db.cursor()

        for sql in TABLE_CREATION:
            self._db_exec(sql, cu=cu, ignore_errors=True)

    def _make_spooledfile_object(self,
                                 uuid,
                                 message,
                                 state,
                                ):
        result = SpooledFile.__new__(SpooledFile)
        result._spool = self
        result._uuid = uuid
        result._message = message
        return result

    def add_incoming(self,
                     message,
                     starting_state = 'U',
                     ):
        assert isinstance(message, str)
        assert starting_state in STATES

        our_uuid = str(uuid.uuid4())

        self._db_exec(
                'insert into message (uuid, content, state) values (?, ?, ?)',
                (
                    our_uuid,
                    message,
                    starting_state,
                    ),
                )
        return self._make_spooledfile_object(
                uuid = our_uuid,
                message = message,
                state = starting_state,
                )

    def items(self):
        for uuid, content, state in self._db_exec(
                'select uuid, content, state from message'
                ):
            yield (uuid,
                   self._make_spooledfile_object(
                       uuid = uuid,
                       message = content,
                       state = state,
                       ))

    def __repr__(self):
        return f'[spool;path={self.path}]'

class SpooledFile:
    def __init__(self):
        raise ValueError(
                "Don't instantiate this."
                )

    def __eq__(self, other):
        if isinstance(other, str):
            return self._uuid == other
        elif isinstance(other, SpooledFile):
            return self._uuid == other._uuid
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    @property
    def spool(self):
        return self._spool

    @property
    def message(self):
        return self._message

    @property
    def state(self):
        found = self._spool._db_exec(
                'select state from message where uuid=?',
                (self._uuid,),
                ).fetchone()
        return found[0]

    @state.setter
    def state(self, v):
        if v not in STATES:
            raise ValueError(f"{v} is not a state (they are {STATES})")
        assert v in STATES
        self._spool._db_exec(
                'update message set state=? where uuid=?',
                (v, self._uuid,),
                )

    @property
    def uuid(self):
        return self._uuid

    def __repr__(self):
        return f'[spooled file {self.uuid}]'
