# fetch.py
#
# Part of kepi.
# Copyright (c) 2018-2020 Marnanel Thurman.
# Licensed under the GNU Public License v2.

import logging
logger = logging.getLogger(name="kepi")

from urllib.parse import urlparse
from requests_cache import (
        CachedSession,
        )
import requests

# XXX generalise this
CACHED_SESSION_PATH = '/tmp/kepi-cache.sqlite3'

# Somewhat based on sombrero. Will merge things when the dust settles.

class Fetch:
    session = None

    @classmethod
    def _get_session(cls):
        cls.session = CachedSession(
            cache_path = CACHED_SESSION_PATH,
                )

    @classmethod
    def fetch(cls,
              address,
              wanted_type = None,
              ):

        """
        Find remote or local objects.

        Args:
            address: the address of the thing we're looking for.
                It's usually a URL. For Persons it can also be atstyle
                ("username@hostname"), which will result in a webfinger
                lookup to find the actual URL. URLs should not contain
                an "@" sign.

            wanted_type (str or None): if this is not None, we check the
                "type" field of the object we retrieve. If it doesn't
                match wanted_type, we log an error and return None.

        Returns:
            dict

        To do:
            If the address refers to localhost, we might avoid doing
                the request.
        """

        result = {
                'success': False,
                }

        if address is None:
            result['problem'] = 'no-address'
            return result

        if cls.session is None:
            cls._get_session()

        wanted = cls._parse_address(address)

        if wanted['is_atstyle']:

            assert False, "not yet implemented"

            webfinger = get_webfinger(
                    username = wanted['username'],
                    hostname = wanted['hostname'],
                    )

            if webfinger.url is None:
                logger.info("webfinger lookup failed; bailing")
                result['problem'] = 'webfinger-failed'
                return result

            logger.debug("webfinger gave us %s", webfinger.url)
            address = webfinger.url

        try:
            response = cls.session.get(
                    address,
                    headers = {
                        'Accept': 'application/activity+json',
                        },
                    )
        except ConnectionError:

            logger.info("%s: cannot reach host",
                address)
            result['problem'] = 'cannot-reach-host'

            return result

        except requests.Timeout:

            logger.info("%s: timeout reaching host",
                address)


            result['problem'] = 'timeout'
            return result

        # so, we have *something*...
        result['status-code'] = response.status_code

        if response.status_code!=200:
            # HTTP error; bail immediately

            logger.info("Unexpected status code from status lookup: %d",
                    response.status_code,
                    )

            result['problem'] = 'status-code'
            return result

        try:
            found = response.json()
        except ValueError as ve:
            logger.info("%s: response was not JSON (%s)",
                    address, response.content)

            result['problem'] = 'not-json'
            return result

        logger.debug("Retrieved: %s", found)
        found |= result
        result = found

        if 'type' not in found:
            logger.info("retrieved JSON did not include a type")

            result['problem'] = 'no-type'
            return result

        if 'id' in found:
            if found['id'] != address:
                logger.info(
                        "user's id was not the source url: got %s",
                        found['id'],
                        )
                result['problem'] = 'wrong-id'
                return result

        result['type'] = found.get('type', None)
        if wanted_type is not None and result['type']!=wanted_type:
            logger.info("Type was %s, but we wanted %s",
                        result['type'], wanted_type)

            result['problem'] = 'wrong-type'
            return result

        result['success'] = True
        return result

    @classmethod
    def turn_caching_off(cls):
        from requests import Session
        cls.session = Session()

    @classmethod
    def _parse_address(cls, address):

        result = {
            'is_atstyle': '@' in address,
        }

        if result['is_atstyle']:
            fields = address.split('@')
            result['username'] = fields[-2]
            result['hostname'] = fields[-1]
        else:
            parsed = urlparse(address)
            result['hostname'] = parsed.netloc
            result['path'] = parsed.path

        logger.debug("%s: wanted: %s", address, result)

        return result

fetch = Fetch.fetch
turn_caching_off = Fetch.turn_caching_off
