# Many of these tests are based on tests from busby, and we should
# refactor those when the dust settles.

from test import *
import logging
import email
import json
from kepi.fastcgi import despatch
from kepi.config import config

logger = logging.getLogger(name='kepi')

HTML_MIMETYPE = 'text/html'

HOST_META_URI       = '/.well-known/host-meta'
HOST_META_MIMETYPE = 'application/xrd+xml'

NODEINFO_PART_1_URI = '/.well-known/nodeinfo'
NODEINFO_PART_2_URI = '/nodeinfo.json'
NODEINFO_MIMETYPE   = (
        'application/json; '
        'profile=http://nodeinfo.diaspora.software/ns/schema/2.0#'
        )

WEBFINGER_BASE_URI = '/.well-known/webfinger'
WEBFINGER_URI = WEBFINGER_BASE_URI + '?resource={}'
WEBFINGER_MIMETYPE = 'application/jrd+json; charset=utf-8'

ACTIVITY_MIMETYPE = 'application/activity+json'

def call_despatch_and_parse_result(*args, **kwargs):
    found = despatch(*args, **kwargs).decode('UTF-8')
    result = email.message_from_string(found)
    return result

def test_fastcgi_simple():
    found = call_despatch_and_parse_result(
            env = {
                'DOCUMENT_URI': '/this-is-not-a-real-URI',
                },
            )
    assert found['Status'].startswith('404 ')

def test_fastcgi_host_meta():

    found = call_despatch_and_parse_result(
            env = {
                'DOCUMENT_URI': HOST_META_URI,
                'ACCEPT': HOST_META_MIMETYPE,
                'SERVER_NAME': 'wombles.example.org',
                },
            )
    assert found['Status'].startswith('200 ')
    assert found['Content-Type']==HOST_META_MIMETYPE

    assert (
        # Not an f-string; the curly brackets are part of the string
        "https://wombles.example.org/.well-known/webfinger?resource={uri}"
        ) in found.get_payload()

def test_fastcgi_nodeinfo_part_1():

    found = call_despatch_and_parse_result(
            env = {
                'DOCUMENT_URI': NODEINFO_PART_1_URI,
                'ACCEPT': NODEINFO_MIMETYPE,
                'SERVER_NAME': 'wombles.example.org',
                },
            )

    assert found['Status'].startswith('200 ')
    assert found['Content-Type']==NODEINFO_MIMETYPE

    response = json.loads(found.get_payload())

    assert 'links' in response

    assert response['links'] == [
            {
                "rel": "http://nodeinfo.diaspora.software/ns/schema/2.0",
                "href": "https://wombles.example.org"+NODEINFO_PART_2_URI,
                },
            ]

def test_fastcgi_nodeinfo_part_2():

    found = call_despatch_and_parse_result(
            env = {
                'DOCUMENT_URI': NODEINFO_PART_2_URI,
                'ACCEPT': NODEINFO_MIMETYPE,
                'SERVER_NAME': 'wombles.example.org',
                },
            )

    assert found['Status'].startswith('200 ')
    assert found['Content-Type']==NODEINFO_MIMETYPE

    response = json.loads(found.get_payload())

    assert response['version']=='2.0'
    assert response['software']['name']=='kepi'
    assert 'activitypub' in response['protocols']

"""
class _TestUsers:

    def __enter__(self):
        self.tempd = tempfile.TemporaryDirectory()
        print("9444", "================", self.tempd)
        config.users_dir = self.tempd.name
        print("9445", "================", config.settings)
        return self.tempd.__enter__()

    def __exit__(self, exc_type, exc_value, exc_tb):
        return self.tempd.__exit__(exc_type, exc_value, exc_tb)
        """

def test_fastcgi_webfinger():

    def webfinger(who,
                  expected_status = 200,
                  ):
        uri = WEBFINGER_BASE_URI
        if who is not None:
            who = f'resource={who}'

        found = call_despatch_and_parse_result(
                env = {
                    'DOCUMENT_URI': WEBFINGER_BASE_URI,
                    'ACCEPT': WEBFINGER_MIMETYPE,
                    'QUERY_STRING': who,
                    'SERVER_NAME': 'wombles.example.org',
                    },
                )

        assert found['Status'].startswith(
                 f'{expected_status} '), 'Status is correct'

        return found

    config.authorise_follow_uri = '/authorise-follow'
    config.user_uri = '/users/%(username)s'

    config.users_dir = os.path.join(
            os.path.dirname(__file__),
            'example-users',
            )

    response = webfinger(
            who = None,
            expected_status = 400,
            )

    response = webfinger(
            who = 'I like coffee',
            expected_status = 400,
            )

    response = webfinger(
            who = 'jamie@magic-torch.example.net',
            expected_status = 404,
            )

    response = webfinger(
            who = 'lord_lucan@wombles.example.org',
            expected_status = 404,
            )

    response = webfinger(
            who = 'sheila@wombles.example.org',
            expected_status = 200,
            )

    assert response['Content-Type']==WEBFINGER_MIMETYPE

    assert response['Access-Control-Allow-Origin']=='*', (
            'ACAO is *, per RFC'
            )

    parsed = json.loads(response.get_payload())

    assert parsed['subject']=='acct:sheila@wombles.example.org'
    assert 'https://wombles.example.org/users/sheila' in parsed['aliases']
    assert {
                'rel': 'self',
                'type': 'application/activity+json',
                'href': 'https://wombles.example.org/users/sheila',
                } in parsed['links']

def test_fastcgi_user():
    found = call_despatch_and_parse_result(
            env = {
                'DOCUMENT_URI': '/users/alice',
                'ACCEPT': ACTIVITY_MIMETYPE,
                'SERVER_NAME': 'wombles.example.org',
                },
            )
    assert found['Status'].startswith('200 ')
    assert found == 'womle'
