import json
import httpsig
import requests_mock
from kepi.validate import validate
import kepi.fetch

# Much of this is taken from the bowler_pub tests; we're keeping them 
# separate for now because they use different libraries, but when
# things calm down a bit we'll put them back together.
ACTIVITY_ID = "https://example.com/04b065f8-81c4-408e-bec3-9fb1f7c06408"
ACTIVITY_DATE = "Thu, 04 Apr 2019 21:12:11 GMT"
INBOX_HOST = 'europa.example.com'
INBOX_PATH = '/sharedInbox'

REMOTE_FRED = 'https://remote.example.org/users/fred'
REMOTE_JIM = 'https://remote.example.org/users/jim'

FREDS_INBOX = REMOTE_FRED+'/inbox'
JIMS_INBOX = REMOTE_JIM+'/inbox'
REMOTE_SHARED_INBOX = 'https://remote.example.org/shared-inbox'

LOCAL_ALICE = 'https://testserver/users/alice'
LOCAL_BOB = 'https://testserver/users/bob'

PUBLIC = "https://www.w3.org/ns/activitystreams#Public"
CONTEXT_URL = "https://www.w3.org/ns/activitystreams"

def message_for_testing(
        secret='',
        path=INBOX_PATH,
        host=INBOX_HOST,
        signed = True,
        fields = {},
        ):

    body = fields.copy()
    body['@context'] = CONTEXT_URL
    body['Host'] = host

    headers = {
            'content-type': "application/activity+json",
            'date': ACTIVITY_DATE,
            'host': host,
            'path': path,
            }

    if 'key_id' in fields:
        key_id = fields['key_id']
    else:
        # the only reason "actor" would be missing from body
        # is if we're deliberately messing around with the body
        # but we still need the headers. In that case it's
        # not a problem for the test value of the key name
        # to be "#main-key".
        key_id = body.get('actor', '')+'#main-key'

    if signed:

        signer = httpsig.HeaderSigner(
                secret=secret,
                algorithm='rsa-sha256',
                key_id = key_id,
                headers=['(request-target)', 'host', 'date', 'content-type'],
                )

        headers = signer.sign(
                headers,
                method='POST',
                path=path,
                )

        SIGNATURE = 'Signature'
        if headers['Authorization'].startswith(SIGNATURE):
            headers['Signature'] = headers['Authorization'][len(SIGNATURE)+1:]

    if 'id' not in body:
        body['id'] = ACTIVITY_ID

    headers['body'] = json.dumps(body,
                                 indent = 2,
                                 sort_keys = True,
                                 )

    return headers

def remote_user(
        remote_url, name,
        public_key='',
        shared_inbox=None,
        ):
    result = {
            '@context': CONTEXT_URL,
            'id': remote_url,
            'type': 'Person',
            'following': remote_url+"/following",
            'followers': remote_url+"/followers",
            'inbox': remote_url+"/inbox",
            'outbox': remote_url+"/outbox",
            'featured': remote_url+"/collections/featured",
            'preferredUsername': name,
            'url': remote_url,
            'publicKey': {
                'id': remote_url+'#main-key',
                'owner': remote_url,
                'publicKeyPem': public_key,
                },
            }

    if shared_inbox is not None:
        result['endpoints'] = {
                'sharedInbox': shared_inbox,
                }

    return json.dumps(result,
                      indent = 2,
                      sort_keys = True,
                      )

def test_validation_passes(requests_mock):

    with open('kepi/bowler_pub/tests/keys/keys-0001.json', 'r') as f:
        keys = json.load(f)

    requests_mock.get(REMOTE_FRED,
                      text=remote_user(
                          remote_url = REMOTE_FRED,
                          name = 'Fred Example',
                          public_key = keys['public'],
                          ))

    msg = message_for_testing(
            fields = {
                'id': ACTIVITY_ID,
                'type': "Follow",
                'actor': REMOTE_FRED,
                'object': LOCAL_ALICE,
                },
            secret = keys['private'],
            path = INBOX_PATH,
            )

    kepi.fetch.turn_caching_off()

    result = validate(message = msg)

    assert result, "validation passed"

def test_validation_gone_general(requests_mock):
    with open('kepi/bowler_pub/tests/keys/keys-0001.json', 'r') as f:
        keys = json.load(f)

    requests_mock.register_uri(
            'GET', REMOTE_FRED,
            status_code = 410)

    msg = message_for_testing(
            fields = {
                'id': ACTIVITY_ID,
                'type': "Follow",
                'actor': REMOTE_FRED,
                'object': LOCAL_ALICE,
                },
            secret = keys['private'],
            path = INBOX_PATH,
            )

    kepi.fetch.turn_caching_off()

    result = validate(message = msg)

    assert not result, "validation failed"

def test_validation_gone_delete_user(requests_mock):
    with open('kepi/bowler_pub/tests/keys/keys-0001.json', 'r') as f:
        keys = json.load(f)

    requests_mock.register_uri(
            'GET', REMOTE_FRED,
            status_code = 410)

    msg = message_for_testing(
            fields = {
                'id': ACTIVITY_ID,
                'type': "Delete",
                'actor': REMOTE_FRED,
                'object': REMOTE_FRED,
                },
            secret = keys['private'],
            path = INBOX_PATH,
            )

    kepi.fetch.turn_caching_off()

    result = validate(message = msg)

    assert result, "validation passed"
